/*This can be mounted */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/buffer_head.h>
#include <linux/blkdev.h>

#include "myfs.h"
#define DEBUG


int myfs_get_super(struct file_system_type *, int,const char *, void *, struct vfsmount *);
void myfs_kill_super(struct super_block * sb);
int myfs_fill_super(struct super_block *, void *, int);
void dump_super_block(struct myfs_super_block *ms);
struct inode * myfs_iget(struct super_block *sb,unsigned long ino);
int myfs_readdir(struct file *filp, void *dirent, filldir_t);
struct dentry * myfs_lookup(struct inode *inode,struct dentry *dentry, struct nameidata *nd);
static int  myfs_get_inode(struct inode* parent_inode,char* name);

static struct inode_operations myfs_inode_ops = {
				.lookup = myfs_lookup
};

static struct file_operations myfs_dir_ops = {
		.read =  generic_read_dir,
		.readdir = myfs_readdir,

};

 
struct super_operations myfs_super_ops={
};


static struct file_system_type myfs = {
	.name = "myfs3",
	.owner = THIS_MODULE,
	.get_sb = myfs_get_super,
	.kill_sb = kill_block_super,
	.fs_flags = FS_REQUIRES_DEV
};

int myfs_readdir(struct file* filp, void* dirent,filldir_t filldir)
{
 	unsigned32 pos = filp->f_pos;
	struct inode *inode = filp->f_dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct buffer_head *bh;
	struct myfs_dir_entry *dir_entry;
	
	printk(KERN_ALERT"MYFS: readdir is entering");	

	if(pos > (inode->i_size - sizeof(struct myfs_dir_entry)))
	return 0;	
	printk(KERN_ALERT"position is correct");	
	if(!(bh = sb_bread(sb,((struct inode_block_number*) inode->i_private)->block_no))) 
	{
	printk(KERN_ALERT"MYFS: block number  could not be read from disk");
	return -1;
	}	
	again:
	pos = filp->f_pos;
	dir_entry=(struct myfs_dir_entry*)(bh->b_data+pos);	
	if(dir_entry->ino <= 0){
	 // printk(KERN_ALERT"inode number <=0 %d",dir_entry->ino);	
	  filp->f_pos = filp->f_pos + sizeof(struct myfs_dir_entry);
	  if(filp->f_pos <= ((inode->i_size) -sizeof(struct myfs_dir_entry))){	
		  goto again;	
		}
	}
	else{
	printk(KERN_ALERT"called filldir dirent with obtained inode number %d name %s ",dir_entry->ino,dir_entry->name);
	filldir(dirent,dir_entry->name,MAX_NAME_LEN,pos,dir_entry->ino,DT_UNKNOWN);
	  filp->f_pos = filp->f_pos + sizeof(struct myfs_dir_entry);
	}
	brelse(bh);
	printk(KERN_ALERT"MYFS: readdir is returning");	
	return 1;

		
}

static int  myfs_get_inode(struct inode* parent_inode,char* name)
{
	struct super_block * sb ;
	int block_no ;
	struct buffer_head* bh;
	struct myfs_dir_entry * tmp_dirent;
	int count_dirent,tmp_count;
	struct myfs_dir_entry final_dirent;

	printk(KERN_ALERT "MYFS[%s]:IMPORTANT Reading data from block number %d for name:inode %s:%lud \n",__FUNCTION__,((struct inode_block_number *)parent_inode->i_private)->block_no,name,parent_inode->i_ino);

	sb = parent_inode->i_sb;
        block_no  = ((struct inode_block_number *)parent_inode->i_private)->block_no;
        if(!(bh = sb_bread(sb,block_no))){ 
	 printk(KERN_ALERT"MYFS: block number %d could not be read \n",block_no);
	 return 0;
	}
	tmp_dirent = (struct myfs_dir_entry *)bh->b_data;	
	count_dirent = (sb->s_blocksize)/sizeof(struct myfs_dir_entry);	
	tmp_count=1;
	while(tmp_count <= count_dirent )
	{
//	  printk(KERN_ALERT"MYFS: READ DIRECTORY ENTRY NAME UNMATCHED %s",tmp_dirent->name);	
	if(strcmp(tmp_dirent->name,name)==0)
	{
	final_dirent.ino = tmp_dirent->ino;
	memcpy(final_dirent.name,tmp_dirent->name,MAX_NAME_LEN);
	final_dirent.len = tmp_dirent->len;
	 brelse(bh);
	 return final_dirent.ino; 		
	}	
	tmp_dirent = tmp_dirent + sizeof(struct myfs_dir_entry);
	tmp_count = tmp_count+1;
	}	
	printk(KERN_ALERT"MYFS: NO DIR Entry matched the required name %s RETURNING Inode NUmber = -1",name);
	return -1;
}


struct dentry * myfs_lookup(struct inode* parent_inode,struct dentry * child_dentry,struct nameidata* nd )
{

	int ino;
	struct inode * reqd_inode;

	printk(KERN_ALERT"MYFS: Trying to Lookup the dentry name %s ",child_dentry->d_name.name);			
	printk(KERN_ALERT"MYFS: superblock is at [0x%p] ",parent_inode->i_sb);		
	/*GET THE INODE NUMBER FOR THAT NAME IN PARENT INODE*/
	ino = myfs_get_inode(parent_inode,(char *)(child_dentry->d_name.name));
//	ino = myfs_find_entry(parent_inode,(char *)(child_dentry->d_name.name));

	if(ino<=0)
	{
	printk(KERN_ALERT"MYFS [%s] NAME[%s] Lookup Failed",__FUNCTION__,child_dentry->d_name.name);
	reqd_inode= NULL;		
	}
	else
	{
		printk(KERN_ALERT"myfs_iget called from lookup function -- for sb %p and inode number obtained %d",parent_inode->i_sb,ino);
	/*GET INODE FOR THIS INODE NUMBER OBTAINED */
		reqd_inode= myfs_iget(parent_inode->i_sb,ino);
		if(!reqd_inode){
			printk(KERN_ALERT"MYFS: READ inode is NULL for inode number [%d]",ino);
			return ERR_PTR(-EINVAL);
		}
	}
	/*ATTACH TH EINODE WTH GIVEN DENTRY AND RETURN DENTRY */
	d_add(child_dentry,reqd_inode);
	printk(KERN_ALERT"MYFS LOOKUP IS RETURNING after attaching dentry");
	return NULL;	
}




/*struct dentry * myfs_lookup(struct inode* parent_inode,struct dentry * child_dentry,struct nameidata* nd )
{

	int ino;
	struct inode * reqd_inode;

	printk(KERN_ALERT"MYFS: Trying to Lookup the dentry name %s ",child_dentry->d_name.name);			
	printk(KERN_ALERT"MYFS: superblock is at [0x%p] ",parent_inode->i_sb);		
	//GET THE INODE NUMBER FOR THAT NAME IN PARENT INODE
	ino = myfs_get_inode(parent_inode,(char *)(child_dentry->d_name.name));
	if(ino<=0)
	{
	printk(KERN_ALERT"NO INODE NUMBER FOR THIS NAME Lookup Failed");
	return ERR_PTR(-1);
	}
	printk(KERN_ALERT"myfs_iget called from lookup function -- for sb %p and inode number %d",parent_inode->i_sb,ino);
	//GET INODE FOR THIS INODE NUMBER OBTAINED 
	reqd_inode= myfs_iget(parent_inode->i_sb,ino);
	if(reqd_inode<=0){
	printk(KERN_ALERT"MYFS: Cannot read inode number [%d]",ino);
	return ERR_PTR(-EINVAL);
	}
	//ATTACH TH EINODE WTH GIVEN DENTRY AND RETURN DENTRY 
	d_add(child_dentry,reqd_inode);
	printk(KERN_ALERT"MYFS LOOKUP IS RETURNING after attaching dentry");
	return child_dentry;	
}*/

struct inode * myfs_iget(struct super_block *sb,unsigned long ino)
{
	struct inode * inode ;
	struct buffer_head *bh;
	struct myfs_inode  *myfs_ino;
	struct inode_block_number *ibn;
	
	inode = iget_locked(sb,ino);
	if(!inode){
	printk(KERN_ALERT"MYFS:Could not get inode by iget_locked fn");
	return ERR_PTR(-ENOMEM);;
	}
	printk(KERN_ALERT"MYFS: inode obtained for inode number %lud ",ino);
	printk(KERN_ALERT "reading inode from disk sb->s_bdev = %p , blocksize = %ld blocknumbr is %d \n ", sb->s_bdev,sb->s_blocksize,MYFS_INODE_BLOCK);
	if(!(bh=(struct buffer_head *)sb_bread(sb,MYFS_INODE_BLOCK))){
		printk(KERN_ALERT"Cannot read inode block %d for inod enumber %ld\n",MYFS_INODE_BLOCK,ino);
		goto free;
	}
	myfs_ino=(struct myfs_inode *)((char *)(bh->b_data)+ ((ino) * sizeof(struct myfs_inode)));

	ibn = kzalloc(sizeof(struct inode_block_number),GFP_KERNEL);
	if(!ibn){
		printk(KERN_ALERT"MYFS[%s]  Not enough memory",__FUNCTION__);
		return ERR_PTR(-1);
	}
	ibn->block_no = myfs_ino->block_no;	

	inode->i_ino  = myfs_ino->ino; // may not be needed 
	inode->i_uid  =myfs_ino->uid;
	inode->i_gid    = myfs_ino->gid;
	inode->i_mode   = myfs_ino->mode;
	inode->i_nlink  = myfs_ino->links;
	inode->i_size   = myfs_ino->size;
	inode->i_private = ibn;


	unlock_new_inode(inode);
 	brelse(bh);
	return inode;	
	free:
	return ERR_PTR(-1);
}


int myfs_fill_super(struct super_block *sb, void *data, int silent)
{
	
	struct myfs_sb_info *msi;
	struct myfs_super_block *ms;

	struct inode *root_inode;
	struct buffer_head *bh,*bi;
	struct myfs_inode  *myfs_ino;
	
	msi = kmalloc(sizeof(struct myfs_sb_info),GFP_KERNEL);
	if(!msi)
	return -ENOMEM;
	sb->s_fs_info = msi;
	sb->s_blocksize=MYFS_DEFAULT_BS;
	memset(msi,0,sizeof(*msi));	
	/*Read the super block*/
	printk(KERN_ALERT "sb->s_bdev = %p , blocksize = %ld \n", sb->s_bdev,sb->s_blocksize );
	if(!(bh=(struct buffer_head *)sb_bread(sb,MYFS_SUPER_BLOCK)))
	{
	  printk(KERN_ALERT"Cannot read superblock of MYFS \n");	
	   goto free;
	}
	else
	printk(KERN_ALERT"Read superblock of MYFS in block [%d] \n",MYFS_SUPER_BLOCK);
	
	ms =(struct myfs_super_block*)(bh->b_data) ;
	msi->ms =ms;
	dump_super_block(ms);
	sb->s_magic = ms->s_magic;
	printk(KERN_ALERT"MYFS: Got the Magic Number [0x%lud]",sb->s_magic);
	if(sb->s_magic != MYFS_SUPER_MAGIC){
		printk(KERN_ALERT"Not a Myfs FIlesystem");
		goto free;
	}
	sb->s_op = &myfs_super_ops;
		

	/*------------READ THE INODE ------------*/
	root_inode=myfs_iget(sb,MYFS_ROOT_INO);
	printk(KERN_ALERT"MYFS: root inode is read by myfs_iget ");

	/*give inode number to already fully initialized super block -- to get root inode*/

	sb->s_root = d_alloc_root(root_inode);
	if(!sb->s_root){
		iput(root_inode);
		printk(KERN_ALERT"MYFS: Cannot read root inode");	
		goto free;
	}	
	 	printk(KERN_ALERT"MYFS: dentry is attached  ");
	if(!S_ISDIR(root_inode->i_mode)){
		dput(sb->s_root);
		sb->s_root = NULL;
		printk(KERN_ALERT"MYFS:Error reading root inode");
		goto free;
	}	 
	root_inode->i_fop = &myfs_dir_ops;
	root_inode->i_op = &myfs_inode_ops;
			
	 	printk(KERN_ALERT"MYFS: root inode's mode is checked successfully  ");
	return 0;
	free:
	 	kfree(msi);	
		return -EINVAL;	 		
}


void dump_super_block(struct myfs_super_block *ms)
{
printk(KERN_ALERT"dumping super block of MYFS : root %d,block_bmp = %d,inode_bmp =%d,block_size = %d,blocks = %d,s_magic = 0x%x,magic %s \n",ms->root,ms->block_bmp,ms->inode_bmp,ms->block_size,ms->blocks,ms->s_magic,ms->magic);
}



int myfs_get_super(struct file_system_type *fs, int flags,const char *dev_name , void * data , struct vfsmount *mnt)
{
 return get_sb_bdev(fs,flags,dev_name,data,myfs_fill_super,mnt);
}
/*void myfs_kill_super(struct super_block * sb)
{

return;

}*/



static void myfs_exit(void)
{
	int err=0;
#ifdef DEBUG
	printk(KERN_INFO"Unregistering Filesystem myfs");
#endif
	if( (err=unregister_filesystem(&myfs)) !=0)
	printk(KERN_INFO"Error Unloading filesystem %d ",err);

}


static int myfs_init(void)
{
int err=0;
#ifdef DEBUG
printk(KERN_INFO"Registering Filesystem myfs \n");
#endif
	err = register_filesystem(&myfs);
	return err;
}
module_init(myfs_init);
module_exit(myfs_exit);



