#include<stdio.h>
#include<fcntl.h>
#include<stdlib.h>
#include<sys/stat.h>
#include<unistd.h>
#include<string.h>
#include "myfs.h"

inline int set_myfs_bit(int bitmap,int blockno){
				return (bitmap)|(1<<(blockno-1));
}

int bwrite(int dev, int block, int blocksize,char *buff){
				int nw;
				lseek(dev,block*blocksize,SEEK_SET);
	fprintf(stderr,"\nwriting on dev %d block %d data %s",dev,block,buff);
	

	if((nw=write(dev,buff,blocksize))!=blocksize)
	fprintf(stderr,"Could not write completely to the buffer\nWrote %d bytes out of %d\n",nw,blocksize);
	fprintf(stderr,"\n Number of bytes written %d",nw);
				return nw;
}

struct buffer_head * bread(int dev, int block, int blocksize){
		
				struct buffer_head *bh;
				if((bh=(struct buffer_head *)malloc(sizeof(*bh)))==NULL)
				return NULL;
				bh->b_data=(char *)malloc(blocksize);
				lseek(dev,block*blocksize,SEEK_SET);
				read(dev,bh->b_data,blocksize);
				bh->b_dev=dev;
				bh->b_size=blocksize;
				bh->b_blocknr=block;
				return bh;
}
void brelse(struct buffer_head *bh)
{
	free(bh->b_data);
	free(bh);
}



void create_root_inode(int dev,unsigned32 blockno,unsigned32 inode,unsigned32 blocksize){
				struct buffer_head *bh,*bh1;
				struct myfs_inode *root;
	
			        fprintf(stderr,"Writing root inode [%d] info in block [%d]\n",inode,blockno);
				bh=bread(dev,blockno,blocksize);
				if(!bh){
				fprintf(stderr,"Could not create root inode\n");
				exit(1);
				}
				root=(struct myfs_inode *)((char *)bh->b_data+inode*sizeof(*root));
			        fprintf(stderr,"Writing root inode [%d] info in block [%d]\n",inode,blockno);
					/*Now write the information of root inode*/
				root->ino=inode;
				root->uid=root->gid=0;
				root->mode|=(S_IFDIR|S_IRUSR|S_IWUSR|S_IXUSR|S_IROTH|S_IXOTH|S_IRGRP|S_IXGRP);
				root->perm=0x0755;
				root->size=blocksize;
				root->block_no=blockno;
				root->links=2;
				fprintf(stderr," ");
				bwrite(dev,blockno,blocksize,bh->b_data);
				
				bh1=bread(dev,blockno,blocksize);
				if(!bh1){
				fprintf(stderr,"Could not READ root inode\n");
				exit(1);
				}
				root=(struct myfs_inode *)((char *)bh1->b_data+inode*sizeof(*root));
				fprintf(stderr,"root->block_no %d",root->block_no);	
				fprintf(stderr,"root->links %d",root->links);	
				

				return;
}


void myfs_create(char * filename,unsigned int bs){
	struct myfs_filesys *fs;
	struct myfs_super_block *sb;
	struct buffer_head *bh,*bh1;
	struct myfs_inode *root,*root1;
	char  * inode_addr;
	char * name_str;
	struct myfs_dir_entry mde; 

	if(bs<=0 || bs>4096){
		fprintf(stderr,"Either invalid blocksize or too big for MYFS !!..setting to default [%d]\n",MYFS_DEFAULT_BS);
		bs=MYFS_DEFAULT_BS;
	}
	fs=(struct myfs_filesys *)malloc(sizeof(*fs));
	fs->fs_dev=open(filename,O_RDWR);
	if(fs->fs_dev==-1){
		perror("Cannot open partition ");
		goto free;
	}
	fs->blocksize=bs; /* Set the blocksize */
	fprintf(stderr,"Writing super block info in block [%d]\n",MYFS_SUPER_BLOCK);
	fs->bh=bread(fs->fs_dev,MYFS_SUPER_BLOCK,bs); /* Get buffer of superblock */
	if(!fs->bh){
		fprintf(stderr,"Could not get buffer for superblock \n");
		goto free;
	}
	
	fs->myfs_ondisk_sb = (struct myfs_super_block *)(fs->bh->b_data);
	memset(fs->myfs_ondisk_sb,0,sizeof(struct myfs_super_block));
	fs->myfs_ondisk_sb->root=MYFS_ROOT_INO;
	set_myfs_bit(fs->myfs_ondisk_sb->block_bmp,MYFS_SUPER_BLOCK);
	set_myfs_bit(fs->myfs_ondisk_sb->block_bmp,MYFS_ROOT_BLOCK);
	set_myfs_bit(fs->myfs_ondisk_sb->block_bmp,MYFS_INODE_BLOCK);
	set_myfs_bit(fs->myfs_ondisk_sb->inode_bmp,MYFS_ROOT_INO);

	fs->myfs_ondisk_sb->blocks=MYFS_MAX_BLOCKS;
	fs->myfs_ondisk_sb->s_magic=MYFS_SUPER_MAGIC;
	fs->myfs_ondisk_sb->block_size= bs;
	strcpy((fs->myfs_ondisk_sb->magic),MYFS_MAGIC_STR);
	fs->bh->b_data =(char *)fs->myfs_ondisk_sb; 	
	bwrite(fs->fs_dev,MYFS_SUPER_BLOCK,bs,fs->bh->b_data);
	
	bh = bread(fs->fs_dev,MYFS_SUPER_BLOCK,bs);
	sb=(struct myfs_super_block *)bh->b_data;
	fprintf(stderr,"root inode %d,block_bmp %d , inode bmp %d,blocks %d , magic_str %s",sb->root,sb->block_bmp,sb->inode_bmp,sb->blocks,sb->magic);
	fprintf(stderr,"Going to create root inode %d,inode block number %d,blocksize %d",MYFS_ROOT_INO,MYFS_INODE_BLOCK,fs->blocksize);
	


	/*CREATING ROORT INODE ---*/

	bh=bread(fs->fs_dev,MYFS_INODE_BLOCK,bs);
	if(!bh){
		fprintf(stderr,"Could not create root inode\n");
		exit(1);
	}
	memset(bh->b_data,0,bs);
	root=(struct myfs_inode *)((char *)bh->b_data + (MYFS_ROOT_INO)*(sizeof(struct myfs_inode)));
	root->ino=MYFS_ROOT_INO;
	root->uid=root->gid=0;
	root->mode|=(S_IFDIR|S_IRUSR|S_IWUSR|S_IXUSR|S_IROTH|S_IXOTH|S_IRGRP|S_IXGRP);
	root->perm=0x0755;
	root->size=bs;
	root->block_no = MYFS_ROOT_BLOCK;
	root->links=2;
	
	bwrite(fs->fs_dev,MYFS_INODE_BLOCK,bs,bh->b_data);

	/*TO CHECK WHETHER WRITTEN CORRECTLY OR NOT */
	bh1=bread(fs->fs_dev,MYFS_INODE_BLOCK,bs);
	if(!bh1){
	 	fprintf(stderr,"Could not READ root inode\n");
		exit(1);
	}
     root1=(struct myfs_inode *)((char *)bh1->b_data+MYFS_ROOT_INO*sizeof(*root));
     fprintf(stderr,"root->block_no %d",root1->block_no);	
     fprintf(stderr,"root->links %d",root1->links);
     brelse(bh1);
     brelse(bh)	;

	bh=bread(fs->fs_dev,MYFS_ROOT_BLOCK,bs);
	memset(bh->b_data,0,bs);

	memset(&mde,0,sizeof(struct myfs_dir_entry));
	mde.ino=MYFS_ROOT_INO;
	strcpy(mde.name,".");
	mde.len=1;
	memcpy(bh->b_data,&mde,sizeof(mde));
	
	memset(&mde,0,sizeof(struct myfs_dir_entry));
	mde.ino=MYFS_ROOT_INO;
	strcpy(mde.name,"..");
	mde.len=2;
	memcpy(bh->b_data+sizeof(struct myfs_dir_entry),&mde,sizeof(mde));
	bwrite(fs->fs_dev,MYFS_ROOT_BLOCK,bs,bh->b_data);
	brelse(bh);


//	create_root_inode(fs->fs_dev,MYFS_INODE_BLOCK,MYFS_ROOT_INO,fs->blocksize);
	fprintf(stderr,"MYFS filesystem created successfully\n");
	return;
	free:
		free(fs);
	 return;
}
